/*
 * mm_radio_resource.c
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include "mm_radio_utils.h"
#ifdef TIZEN_FEATURE_RADIO_HAL
#include "mm_radio_priv_hal.h"
#else
#include "mm_radio_priv.h"
#endif
#include "mm_radio_resource.h"
#include <murphy/common/glib-glue.h>

#define MMRADIO_RESOURCE_TIMEOUT 5
#define MRP_APP_CLASS_FOR_RADIO   "media"
#define MRP_RESOURCE_TYPE_MANDATORY TRUE
#define MRP_RESOURCE_TYPE_EXCLUSIVE FALSE

const char* mm_radio_resource_str[MM_RADIO_RESOURCE_MAX] = {
	"radio"
};

#define MMRADIO_CHECK_RESOURCE_MANAGER_INSTANCE(x_radio_resource_manager) \
do { \
	if (!x_radio_resource_manager) { \
		MMRADIO_LOG_ERROR("no resource manager instance"); \
		return MM_ERROR_INVALID_ARGUMENT; \
	} \
} while (0);

#define MMRADIO_CHECK_CONNECTION_RESOURCE_MANAGER(x_radio_resource_manager) \
do { \
	if (!x_radio_resource_manager) { \
		MMRADIO_LOG_ERROR("no resource manager instance"); \
		return MM_ERROR_INVALID_ARGUMENT; \
	} else { \
		if (!x_radio_resource_manager->is_connected) { \
			MMRADIO_LOG_ERROR("not connected to resource server yet"); \
			return MM_ERROR_RESOURCE_NOT_INITIALIZED; \
		} \
	} \
} while (0);

static char *_mmrdio_resource_state_to_str(mrp_res_resource_state_t st)
{
	char *state = "unknown";
	switch (st) {
	case MRP_RES_RESOURCE_ACQUIRED:
		state = "acquired";
		break;
	case MRP_RES_RESOURCE_LOST:
		state = "lost";
		break;
	case MRP_RES_RESOURCE_AVAILABLE:
		state = "available";
		break;
	case MRP_RES_RESOURCE_PENDING:
		state = "pending";
		break;
	case MRP_RES_RESOURCE_ABOUT_TO_LOOSE:
		state = "about to loose";
		break;
	}
	return state;
}

static void __mmradio_resource_state_callback(mrp_res_context_t *context, mrp_res_error_t err, void *user_data)
{

	int i = 0;
	const mrp_res_resource_set_t *rset;
	mrp_res_resource_t *resource;
	mm_radio_t *radio = NULL;

	MMRADIO_LOG_FENTER();

	if (user_data == NULL) {
		MMRADIO_LOG_ERROR(" - user data is null\n");
		return;
	}
	radio = (mm_radio_t *)user_data;
	if (err != MRP_RES_ERROR_NONE) {
		MMRADIO_LOG_ERROR(" - error message received from Murphy, for the radio(%p), err(0x%x)", radio, err);
		return;
	}

	switch (context->state) {
	case MRP_RES_CONNECTED:
		MMRADIO_LOG_DEBUG(" - connected to Murphy");
		if ((rset = mrp_res_list_resources(context)) != NULL) {
			mrp_res_string_array_t *resource_names;
			resource_names = mrp_res_list_resource_names(rset);
			if (!resource_names) {
				MMRADIO_LOG_ERROR(" - no resources available");
				return;
			}
			for (i = 0; i < resource_names->num_strings; i++) {
				resource = mrp_res_get_resource_by_name(rset, resource_names->strings[i]);
				if (resource)
					MMRADIO_LOG_DEBUG(" - available resource: %s", resource->name);
			}
			mrp_res_free_string_array(resource_names);
		}
		radio->resource_manager.is_connected = true;
		break;
	case MRP_RES_DISCONNECTED:
		MMRADIO_LOG_DEBUG(" - disconnected from Murphy");
		if (radio->resource_manager.rset) {
			mrp_res_delete_resource_set(radio->resource_manager.rset);
			radio->resource_manager.rset = NULL;
		}
		mrp_res_destroy(radio->resource_manager.context);
		radio->resource_manager.context = NULL;
		radio->resource_manager.is_connected = false;
		break;
	}

	MMRADIO_LOG_FLEAVE();
}

static void __mmradio_rset_state_callback(mrp_res_context_t *cx, const mrp_res_resource_set_t *rs, void *user_data)
{
	int i = 0;
	mm_radio_t *radio = (mm_radio_t *)user_data;
	mrp_res_resource_t *res;

	MMRADIO_LOG_FENTER();

	MMRADIO_RESOURCE_LOCK(&radio->resource_manager);

	if (!mrp_res_equal_resource_set(rs, radio->resource_manager.rset)) {
		MMRADIO_LOG_WARNING("- resource set(%p) is not same as this radio handle's(%p)", rs, radio->resource_manager.rset);
		MMRADIO_RESOURCE_UNLOCK(&radio->resource_manager);
		return;
	}

	MMRADIO_LOG_DEBUG(" - resource set state of radio(%p) is changed to [%s]",
		radio, _mmrdio_resource_state_to_str(rs->state));
	for (i = 0; i < MM_RADIO_RESOURCE_MAX; i++) {
		res = mrp_res_get_resource_by_name(rs, mm_radio_resource_str[i]);
		if (res == NULL)
			MMRADIO_LOG_WARNING(" -- %s not present in resource set", mm_radio_resource_str[i]);
		else
			MMRADIO_LOG_DEBUG(" -- resource name [%s] -> [%s]", res->name,
			_mmrdio_resource_state_to_str(res->state));
	}

	mrp_res_delete_resource_set(radio->resource_manager.rset);
	radio->resource_manager.rset = mrp_res_copy_resource_set(rs);

	if (rs->state == MRP_RES_RESOURCE_ACQUIRED) {
		MMRADIO_LOG_DEBUG(" - resource set is acquired");
		radio->resource_manager.state = MM_RADIO_RESOURCE_STATE_ACQUIRED;
		MMRADIO_RESOURCE_SIGNAL(&radio->resource_manager);
	} else if ((radio->resource_manager.state >= MM_RADIO_RESOURCE_STATE_ACQUIRED) &&
			   (rs->state == MRP_RES_RESOURCE_AVAILABLE)) {
		MMRADIO_LOG_DEBUG(" - resource set is released");
		radio->resource_manager.state = MM_RADIO_RESOURCE_STATE_PREPARED;
		MMRADIO_RESOURCE_SIGNAL(&radio->resource_manager);

		/* mm radio was unrealized by resource conflict. */
		if (radio->resource_manager.by_rm_cb == true) {
			MMRADIO_LOG_DEBUG(" - delete resource set ");
			if (radio->resource_manager.rset) {
				mrp_res_delete_resource_set(radio->resource_manager.rset);
				radio->resource_manager.rset = NULL;
			}
			radio->resource_manager.state = MM_RADIO_RESOURCE_STATE_INITIALIZED;
			radio->resource_manager.by_rm_cb = false;
		}
	}

	MMRADIO_RESOURCE_UNLOCK(&radio->resource_manager);

	MMRADIO_LOG_FLEAVE();
}

static void __mmradio_mrp_resource_release_cb(mrp_res_context_t *cx, const mrp_res_resource_set_t *rs, void *user_data)
{
	int i = 0;
	int result = MM_ERROR_NONE;
	mm_radio_t *radio = NULL;
	mrp_res_resource_t *res;
	gboolean resource_released = FALSE;

	MMRADIO_LOG_FENTER();

	if (user_data == NULL) {
		MMRADIO_LOG_ERROR("- user_data is null");
		return;
	}

	radio = (mm_radio_t *)user_data;

	if (!mrp_res_equal_resource_set(rs, radio->resource_manager.rset)) {
		MMRADIO_LOG_WARNING("- resource set(%p) is not same as this radio handle's(%p)", rs, radio->resource_manager.rset);
		return;
	}

	MMRADIO_LOG_DEBUG(" - resource set state of radio(%p) is changed to [%s]", radio,
		_mmrdio_resource_state_to_str(rs->state));
	for (i = 0; i < MM_RADIO_RESOURCE_MAX; i++) {
		res = mrp_res_get_resource_by_name(rs, mm_radio_resource_str[i]);
		if (res == NULL) {
			MMRADIO_LOG_WARNING(" -- %s not present in resource set", mm_radio_resource_str[i]);
		} else {
			MMRADIO_LOG_DEBUG(" -- resource name [%s] -> [%s]", res->name,
				_mmrdio_resource_state_to_str(res->state));
			if (res->state == MRP_RES_RESOURCE_ABOUT_TO_LOOSE)
				resource_released = TRUE;
		}
	}

	if (resource_released) {
		MMRADIO_LOG_DEBUG("radio resource conflict so, resource will be freed by _mmradio_stop");

		radio->resource_manager.by_rm_cb = true;

		MMRADIO_CMD_LOCK(radio);
		result = _mmradio_stop(radio);
		if (result != MM_ERROR_NONE)
			MMRADIO_LOG_ERROR("failed to stop radio");
		MMRADIO_CMD_UNLOCK(radio);

	} else {
		MMRADIO_LOG_WARNING("could not find videobin");
	}

	MMRADIO_LOG_FLEAVE();

	return;

}

static int _mmradio_create_rset(mm_radio_resource_manager *resource_manager)
{
	if (resource_manager->rset) {
		MMRADIO_LOG_ERROR(" - resource set was already created");
		return MM_ERROR_RESOURCE_INVALID_STATE;
	}

	resource_manager->rset = mrp_res_create_resource_set(resource_manager->context,
				MRP_APP_CLASS_FOR_RADIO, __mmradio_rset_state_callback,
				(void*)resource_manager->user_data);
	if (resource_manager->rset == NULL) {
		MMRADIO_LOG_ERROR(" - could not create resource set");
		return MM_ERROR_RESOURCE_INTERNAL;
	}

	if (!mrp_res_set_autorelease(TRUE, resource_manager->rset))
		MMRADIO_LOG_WARNING(" - could not set autorelease flag!");

	return MM_ERROR_NONE;
}

static int _mmradio_include_resource(mm_radio_resource_manager *resource_manager, const char *resource_name)
{
	mrp_res_resource_t *resource = NULL;
	resource = mrp_res_create_resource(resource_manager->rset,
				resource_name,
				MRP_RESOURCE_TYPE_MANDATORY,
				MRP_RESOURCE_TYPE_EXCLUSIVE);
	if (resource == NULL) {
		MMRADIO_LOG_ERROR(" - could not include resource[%s]", resource_name);
		return MM_ERROR_RESOURCE_INTERNAL;
	}

	MMRADIO_LOG_DEBUG(" - include resource[%s]", resource_name);

	return MM_ERROR_NONE;
}

static int _mmradio_set_resource_release_cb(mm_radio_resource_manager *resource_manager)
{
	int ret = MM_ERROR_NONE;
	bool mrp_ret = false;
	MMRADIO_LOG_FENTER();

	if (resource_manager->rset) {
		mrp_ret = mrp_res_set_release_callback(resource_manager->rset, __mmradio_mrp_resource_release_cb, resource_manager->user_data);
		if (!mrp_ret) {
			MMRADIO_LOG_ERROR(" - could not set release callback");
			ret = MM_ERROR_RESOURCE_INTERNAL;
		}
	} else {
		MMRADIO_LOG_ERROR(" - resource set is null");
		ret = MM_ERROR_RESOURCE_INVALID_STATE;
	}
	MMRADIO_LOG_FLEAVE();
	return ret;
}


int mmradio_resource_manager_init(mm_radio_resource_manager *resource_manager, void *user_data)
{
	MMRADIO_LOG_FENTER();
	MMRADIO_CHECK_RESOURCE_MANAGER_INSTANCE(resource_manager);

	GMainContext *mrp_ctx = g_main_context_new();
	if (!mrp_ctx) {
		MMRADIO_LOG_ERROR("- could not create main context for resource manager");
		return MM_ERROR_RESOURCE_INTERNAL;
	}

	GMainLoop *mrp_loop = g_main_loop_new(mrp_ctx, TRUE);
	g_main_context_unref(mrp_ctx);
	if (!mrp_loop) {
		MMRADIO_LOG_ERROR("- could not create glib mainloop for resource manager");
		return MM_ERROR_RESOURCE_INTERNAL;
	}

	resource_manager->mloop = mrp_mainloop_glib_get(mrp_loop);
	g_main_loop_unref(mrp_loop);

	if (resource_manager->mloop) {
		resource_manager->context = mrp_res_create(resource_manager->mloop, __mmradio_resource_state_callback, user_data);
		if (resource_manager->context == NULL) {
			MMRADIO_LOG_ERROR(" - could not get context for resource manager");
			mrp_mainloop_destroy(resource_manager->mloop);
			resource_manager->mloop = NULL;
			return MM_ERROR_RESOURCE_INTERNAL;
		}
		resource_manager->user_data = user_data;
	} else {
		MMRADIO_LOG_ERROR("- could not get mainloop for resource manager");
		return MM_ERROR_RESOURCE_INTERNAL;
	}

	resource_manager->state = MM_RADIO_RESOURCE_STATE_INITIALIZED;
	g_mutex_init(&resource_manager->lock);
	g_cond_init(&resource_manager->cond);

	MMRADIO_LOG_FLEAVE();

	return MM_ERROR_NONE;

}


int mmradio_resource_manager_prepare(mm_radio_resource_manager *resource_manager, mm_radio_resource_type_e resource_type)
{
	int ret = MM_ERROR_NONE;
	MMRADIO_LOG_FENTER();
	MMRADIO_CHECK_RESOURCE_MANAGER_INSTANCE(resource_manager);
	MMRADIO_CHECK_CONNECTION_RESOURCE_MANAGER(resource_manager);

	if (!resource_manager->rset)
		ret = _mmradio_create_rset(resource_manager);

	if (ret == MM_ERROR_NONE) {
		switch (resource_type) {
		case MM_RADIO_RESOURCE_TYPE_RADIO:
			ret = _mmradio_include_resource(resource_manager, mm_radio_resource_str[MM_RADIO_RESOURCE_TYPE_RADIO]);
			break;
		default:
			MMRADIO_LOG_WARNING("resource type(%d) is worng", resource_type);
			break;
		}
	}

	resource_manager->state = MM_RADIO_RESOURCE_STATE_PREPARED;

	MMRADIO_LOG_FLEAVE();

	return ret;

}

int mmradio_resource_manager_unprepare(mm_radio_resource_manager *resource_manager)
{
	int ret = MM_ERROR_NONE;
	MMRADIO_LOG_FENTER();
	MMRADIO_CHECK_RESOURCE_MANAGER_INSTANCE(resource_manager);
	MMRADIO_CHECK_CONNECTION_RESOURCE_MANAGER(resource_manager);

	if (resource_manager->rset == NULL) {
		MMRADIO_LOG_ERROR("- could not unprepare for resource_manager, mmradio_resource_manager_prepare() first");
		ret = MM_ERROR_RESOURCE_INVALID_STATE;
	} else {
		MMRADIO_RESOURCE_LOCK(resource_manager);
		mrp_res_delete_resource_set(resource_manager->rset);
		resource_manager->rset = NULL;
		MMRADIO_RESOURCE_UNLOCK(resource_manager);
	}

	resource_manager->state = MM_RADIO_RESOURCE_STATE_INITIALIZED;

	MMRADIO_LOG_FLEAVE();

	return ret;

}

int mmradio_resource_manager_acquire(mm_radio_resource_manager *resource_manager)
{
	int ret = MM_ERROR_NONE;
	MMRADIO_LOG_FENTER();
	MMRADIO_CHECK_RESOURCE_MANAGER_INSTANCE(resource_manager);
	MMRADIO_CHECK_CONNECTION_RESOURCE_MANAGER(resource_manager);

	if (resource_manager->rset == NULL) {
		MMRADIO_LOG_ERROR("- could not acquire resource, resource set is null\n");
		ret = MM_ERROR_RESOURCE_INVALID_STATE;
	} else {
		ret = _mmradio_set_resource_release_cb(resource_manager);
		if (ret) {
			MMRADIO_LOG_ERROR("- could not set resource release cb, ret(0x%x)", ret);
			ret = MM_ERROR_RESOURCE_INTERNAL;
		} else {
			MMRADIO_RESOURCE_LOCK(resource_manager);

			ret = mrp_res_acquire_resource_set(resource_manager->rset);
			if (ret) {
				MMRADIO_LOG_ERROR("- could not acquire resource, ret(%d)", ret);
				ret = MM_ERROR_RESOURCE_INTERNAL;
			} else {
				gint64 end_time = g_get_monotonic_time() + MMRADIO_RESOURCE_TIMEOUT * G_TIME_SPAN_SECOND;

				MMRADIO_LOG_DEBUG("- acquire resource waiting..%p till %lld", resource_manager, end_time);

				if (!MMRADIO_RESOURCE_WAIT_UNTIL(resource_manager, end_time)) {
					MMRADIO_LOG_ERROR("- could not acquire resource");
					ret = MM_ERROR_RESOURCE_INTERNAL;
				} else {
					MMRADIO_LOG_DEBUG("- resources are acquired");
				}

			}
			MMRADIO_RESOURCE_UNLOCK(resource_manager);
		}
	}

	MMRADIO_LOG_FLEAVE();

	return ret;

}

int mmradio_resource_manager_release(mm_radio_resource_manager *resource_manager)
{
	int ret = MM_ERROR_NONE;
	MMRADIO_LOG_FENTER();
	MMRADIO_CHECK_RESOURCE_MANAGER_INSTANCE(resource_manager);
	MMRADIO_CHECK_CONNECTION_RESOURCE_MANAGER(resource_manager);

	if (resource_manager->rset == NULL) {
		MMRADIO_LOG_ERROR("- could not release resource, resource set is null");
		ret = MM_ERROR_RESOURCE_INVALID_STATE;
	} else {
		if (resource_manager->rset->state != MRP_RES_RESOURCE_ACQUIRED) {
			MMRADIO_LOG_ERROR("- could not release resource, resource set state is [%s]",
				_mmrdio_resource_state_to_str(resource_manager->rset->state));
			ret = MM_ERROR_RESOURCE_INVALID_STATE;
		} else {
			MMRADIO_RESOURCE_LOCK(resource_manager);
			ret = mrp_res_release_resource_set(resource_manager->rset);
			if (ret) {
				MMRADIO_LOG_ERROR("- could not release resource, ret(%d)", ret);
				ret = MM_ERROR_RESOURCE_INTERNAL;
			} else {
				gint64 end_time = g_get_monotonic_time() + MMRADIO_RESOURCE_TIMEOUT * G_TIME_SPAN_SECOND;

				MMRADIO_LOG_DEBUG("- release resource waiting..%p till %lld", resource_manager, end_time);

				if (!MMRADIO_RESOURCE_WAIT_UNTIL(resource_manager, end_time))
					MMRADIO_LOG_WARNING("- could not release resource in time");
				else
					MMRADIO_LOG_DEBUG("- resources are released\n");
			}
			MMRADIO_RESOURCE_UNLOCK(resource_manager);
		}
	}

	MMRADIO_LOG_FLEAVE();

	return ret;
}

int mmradio_resource_manager_deinit(mm_radio_resource_manager *resource_manager)
{
	MMRADIO_LOG_FENTER();
	MMRADIO_CHECK_RESOURCE_MANAGER_INSTANCE(resource_manager);
	MMRADIO_CHECK_CONNECTION_RESOURCE_MANAGER(resource_manager);

	if (resource_manager->rset) {
		if (resource_manager->rset->state == MRP_RES_RESOURCE_ACQUIRED) {
			if (mrp_res_release_resource_set(resource_manager->rset))
				MMRADIO_LOG_ERROR("- could not release resource");
		}
		mrp_res_delete_resource_set(resource_manager->rset);
		resource_manager->rset = NULL;
	}

	if (resource_manager->context) {
		mrp_res_destroy(resource_manager->context);
		resource_manager->context = NULL;
	}

	if (resource_manager->mloop) {
		mrp_mainloop_quit(resource_manager->mloop, 0);
		mrp_mainloop_destroy(resource_manager->mloop);
		resource_manager->mloop = NULL;
	}

	resource_manager->state = MM_RADIO_RESOURCE_STATE_NONE;
	g_mutex_clear(&resource_manager->lock);
	g_cond_clear(&resource_manager->cond);

	MMRADIO_LOG_FLEAVE();

	return MM_ERROR_NONE;
}


int mmradio_resource_manager_get_state(mm_radio_resource_manager *resource_manager, mm_radio_resource_state_e *state)
{

	MMRADIO_LOG_FENTER();
	MMRADIO_CHECK_RESOURCE_MANAGER_INSTANCE(resource_manager);
	MMRADIO_CHECK_ARG(state);

	MMRADIO_LOG_DEBUG("resource_state is %d", resource_manager->state);

	*state = resource_manager->state;

	MMRADIO_LOG_FLEAVE();

	return MM_ERROR_NONE;
}
